#include <xc.h>
#include "can.h"

#include "pwm.h"
#include "led.h"

#define BUFFERCOUNT 8
static unsigned int CANMessageBuffer[BUFFERCOUNT][8]
__attribute__ ((aligned(BUFFERCOUNT * 16)));

void canInit(unsigned short ownID)
{
    // To set configuration, the ECAN module must be in configuration mode.
    // This cannot be set directly, but must be requested and waited for.
    C1CTRL1bits.REQOP = 4;
    while(C1CTRL1bits.OPMODE != 4);
    
    // Configuration can be set now, first part is the bitrate
    // To get from 70MHz (F_CAN, half of the peripheral clock) to 500kHz we need
    // an overall divider of 140. A CAN bit is divided into 8-25 TQ, and the
    // prescaler is configurable from 1 to 64.
    // One nice and clean division is a CAN bit of 20 TQ and a prescaler of 7
    // Another possibility is 14 TQ and a prescaler of 10, but I won't do that
    C1CFG1bits.BRP = 6; // set prescaler to 7
    C1CFG1bits.SJW = 1; // set sync phase to 2 TQ (standard says 1, should be more robust with 2?)
    C1CFG2bits.PRSEG = 5; // set propagation phase to 6 TQ
    C1CFG2bits.SEG1PH = 5; // set phase 1 to 6 TQ
    C1CFG2bits.SEG2PHTS = 1; // phase 2 can be freely programmed
    C1CFG2bits.SEG2PH = 5; // set phase 2 to 6 TQ
    C1CFG2bits.SAM = 0; // bus is sampled once at the sample point (why not 3 times? TODO: check!)
    
    // configure message buffers in RAM
    C1FCTRLbits.DMABS = 2; // there are 8 buffers
    C1FCTRLbits.FSA = 7; // not using the FIFO, so it points to the last buffer that exists
    
    // Set the message filters and filter masks
    // This is, for now, specific to the motorboard...
    // To access the filter configuration, the WIN bit in CiCTRL1 has to be set:
    C1CTRL1bits.WIN = 1;
    
    // Now, we set up four message filters. Only messages passing those will
    // be received!
    C1FEN1 = 0; // Someone is very stupid, so by default all filters are active
    // and we have to first disable all of them...
    // Filter 0 for clock syncing messages (ID 0x0001) and emergency stop
    // messages (ID 0x0001) -> buffer 4
    C1FMSKSEL1bits.F0MSK = 0; // Filter 0 will use mask 0
    C1RXM0SIDbits.SID = 0x07FF; // All filter bits are active
    C1RXF0SIDbits.SID = 0x0001; // Filter filters for SID 0x0001
    C1RXM0SIDbits.MIDE = 1; // Use this filter on standard identifiers
    C1RXF0SIDbits.EXIDE = 0; // Do not use this filter on extended identifiers
    C1BUFPNT1bits.F0BP = 4; // Messages fitting filter 0 go into buffer 4
    C1FEN1bits.FLTEN0 = 1; // Filter is enabled
    // Filter 1 for clock syncing messages (OD 0x0002) -> buffer 5
    C1FMSKSEL1bits.F1MSK = 0; // Filter 1 will use mask 0
    C1RXF1SIDbits.SID = 0x0002; // Filter filters for SID 0x0002
    C1RXF1SIDbits.EXIDE = 0; // Do not use this filter on extended identifiers
    C1BUFPNT1bits.F1BP = 5; // Messages fitting filter 1 go into buffer 5
    C1FEN1bits.FLTEN1 = 1; // Filter is enabled
    // Filter 2 for axis speed commands (ID 0x001X) -> buffer 6
    C1FMSKSEL1bits.F2MSK = 0; // Filter 2 will use mask 0
    C1RXF2SIDbits.SID = 0x0010 | (unsigned int) ownID; // filters for SID 0x001X
    C1RXF2SIDbits.EXIDE = 0; // Do not use this filter on extended identifiers
    C1BUFPNT1bits.F2BP = 6; // Messages fitting filter 2 go into buffer 6
    C1FEN1bits.FLTEN2 = 1; // Filter is enabled
    // Filter 3 for axis configuration commands (ID 0x002X) -> buffer 7
    C1FMSKSEL1bits.F3MSK = 0; // Filter 3 will use mask 0
    C1RXF3SIDbits.SID = 0x0020 | (unsigned int) ownID; // filters for SID 0x001X
    C1RXF3SIDbits.EXIDE = 0; // Do not use this filter on extended identifiers
    C1BUFPNT1bits.F3BP = 7; // Messages fitting filter 3 go into buffer 7
    C1FEN1bits.FLTEN3 = 1; // Filter is enabled
    
    // Before doing anything else, register access has to be switched back:
    C1CTRL1bits.WIN = 0;
    
    // Set up and activate DMA0 for Tx
    DMA0CONbits.SIZE = 0;
    DMA0CONbits.DIR = 1;
    DMA0CONbits.AMODE = 2;
    DMA0CONbits.MODE = 0;
    DMA0REQ = 70;
    DMA0PAD = (volatile unsigned int) &C1TXD;
    DMA0STAL = (unsigned int) &CANMessageBuffer;
    DMA0STAH = (unsigned int) &CANMessageBuffer;
    DMA0CONbits.CHEN = 1;
    
    // Set up and activate DMA1 for Rx
    DMA1CONbits.SIZE = 0;
    DMA1CONbits.DIR = 0;
    DMA1CONbits.AMODE = 2;
    DMA1CONbits.MODE = 0;
    DMA1REQ = 34;
    DMA1CNT = 7;
    DMA1PAD = (volatile unsigned int) &C1RXD;
    DMA1STAL = (unsigned int) &CANMessageBuffer;
    DMA1STAH = (unsigned int) &CANMessageBuffer;
    DMA1CONbits.CHEN = 0x1;
    //IEC0bits.DMA1IE = 1; // enable interrupt
    
    // Activate the Rx interrupt
    C1INTEbits.RBIE = 1;
    IEC2bits.C1IE = 1;
    
    // Configure Message Buffers 0 to 3 for Transmit
    C1TR01CONbits.TXEN0 = 1;
    C1TR01CONbits.TX0PRI = 3;
    C1TR01CONbits.TXEN1 = 1;
    C1TR01CONbits.TX1PRI = 3;
    C1TR23CONbits.TXEN2 = 1;
    C1TR23CONbits.TX2PRI = 3;
    C1TR23CONbits.TXEN3 = 1;
    C1TR23CONbits.TX3PRI = 3;
    
    // The module must be requested to switch to operation mode, then we must
    // wait for it to do so:
    C1CTRL1bits.REQOP = 0;
    while(C1CTRL1bits.OPMODE != 0);
}

int canSendMessage(unsigned char buffer, CAN_MESSAGE_T* message)
{
    int i;
    
    switch(buffer)
    {
        case 0:
            if(!C1TR01CONbits.TXREQ0)
            {
                for(i=0;i<8;i++)
                {
                    CANMessageBuffer[0][i] = *((unsigned int*) message + i);
                }
                C1TR01CONbits.TXREQ0 = 1;
                return 0;
            }
            return -1; // buffer still in use
        case 1:
            if(!C1TR01CONbits.TXREQ1)
            {
                for(i=0;i<8;i++)
                {
                    CANMessageBuffer[1][i] = *((unsigned int*) message + i);
                }
                C1TR01CONbits.TXREQ1 = 1;
                return 0;
            }
            return -1; // buffer still in use
        case 2:
            if(!C1TR23CONbits.TXREQ2)
            {
                for(i=0;i<8;i++)
                {
                    CANMessageBuffer[2][i] = *((unsigned int*) message + i);
                }
                C1TR23CONbits.TXREQ2 = 1;
                return 0;
            }
            return -1; // buffer still in use
        case 3:
            if(!C1TR23CONbits.TXREQ3)
            {
                for(i=0;i<8;i++)
                {
                    CANMessageBuffer[3][i] = *((unsigned int*) message + i);
                }
                C1TR23CONbits.TXREQ3 = 1;
                return 0;
            }
            return -1; // buffer still in use
    }
    return -1; // buffer doesn't exist
}

void __attribute__ ( (interrupt, no_auto_psv) ) _C1Interrupt( void )
{
    IFS2bits.C1IF = 0;      // clear interrupt flag
    if( C1INTFbits.TBIF )
    {
        C1INTFbits.TBIF = 0;
    }
    
    C1INTFbits.RBOVIF = 0;

    if( C1INTFbits.RBIF )
    {
        if(C1RXFUL1bits.RXFUL4)
        {
            C1RXFUL1bits.RXFUL4 = 0;
        }
        if(C1RXFUL1bits.RXFUL5)
        {
            C1RXFUL1bits.RXFUL5 = 0;
        }
        if(C1RXFUL1bits.RXFUL6)
        {
            if(CANMessageBuffer[6][3]) LATBbits.LATB14 = 1;
            else LATBbits.LATB14 = 0;
            pwmSetDutyCycle(CANMessageBuffer[6][4]);
            if(CANMessageBuffer[6][4]) setLED(RED, ON);
            else setLED(RED, OFF);
            // reset message buffer
            C1RXFUL1bits.RXFUL6 = 0;
        }	    	    
        if(C1RXFUL1bits.RXFUL7)
        {
            C1RXFUL1bits.RXFUL7 = 0;
        }
        C1INTFbits.RBIF = 0;
    }
}
