/*
 * File:   motorboard.c
 * Author: toffi
 *
 * Created on April 16, 2018, 11:52 PM
 */


#include <xc.h>
#include "pic_config.h"
#include "sw_config.h"
#include "led.h"
#include "pwm.h"
#include "can.h"
#include "qei.h"
#include "uart.h"

int main(void)
{
    // ========================================================================
    // Step 0: Oscillator Switch to get 140 MHz / 70 MIPS
    // No error handling here, if the oscillator fails we're fucked anyway
    // We have an 8MHz resonator, so we transform 8 -> 4 -> 280 -> 140 MHz
    PLLFBD = 68; // set PLL divider to 70 (from 4 to 280 MHz)
    CLKDIV = 0; // set N1 and N2 to 2 (8 to 4 and 280 to 140 MHz)
    __builtin_write_OSCCONH(0x03); // prepare clock switch to XTPLL
    __builtin_write_OSCCONL(OSCCON | 0x01); // initiate clock switch
    while(OSCCONbits.COSC != 0b011); // wait for XT to stabilize
    while(OSCCONbits.LOCK != 1); // wait for PLL lock
    // If we make it until here, the PIC is running at 70 MIPS.
    
    // ========================================================================
    // Step 1: I/O configuration
    // Normal Port Pin Configuration
    ANSELA = 0x0001; // AN0 is used for motor current measurement, all others off
    TRISA = 0xFFFF;
    TRISAbits.TRISA1 = 0; // RA1 is an output (LED red)
    
    ANSELB = 0; // no analog functions of port B used
    LATB = 0x0080; // ~CS of the encoder is high
    TRISB = 0xFFFF;
    TRISBbits.TRISB0 = 0; // RB0 (LED green)
    TRISBbits.TRISB1 = 0; // RB1 (LED blue)
    TRISBbits.TRISB7 = 0; // RB7 (Encoder ~CS
    TRISBbits.TRISB13 = 0; // RB13 (Motor PWM)
    TRISBbits.TRISB14 = 0; // RB14 (Motor direction)
    
    // Peripheral Pin Select Configuration
    // unlock PPS registers
    __builtin_write_OSCCONL(OSCCON & ~(1<<6));
    // CAN1 Module
    RPINR26bits.C1RXR = 36; // CAN Rx is mapped to RP36 / Pin RB4
    RPOR0bits.RP20R = 0x0E; // RP20 / Pin RA4 mapped to CAN Tx
    // QEI module
    RPINR14bits.QEA1R = 37; // assign pin RB5 / RP37 to QEI A 
    RPINR14bits.QEB1R = 42; // assign pin RB10 / RP42 to QEI B 
    RPINR15bits.INDX1R = 41; // assign pin RB9 / RP41 to INDEX 
    
    // UART1 module for networking / communications assignment
    RPINR18bits.U1RXR = 44; // Rx on RPI44 / RB12 -> J4
    RPOR4bits.RP43R = 1; // Tx on RP42 / RB11 -> J5
    
    // lock PPS registers
    __builtin_write_OSCCONL(OSCCON | (1<<6));
    
    // ========================================================================
    // Step 2: PWM module configuration
    pwmInit(0xFFFF, 0); // for now, this is a really slow PWM at 1.8kHz, but resolution is great!
    
    // ========================================================================
    // Step 3: ADC configuration
    
    // ========================================================================
    // Step 4: Encoder interface configuration
    qeiInit();
    
    // ========================================================================
    // Step 5: CAN module configuration
    canInit(OWN_AXIS_NUMBER);
    
    // ========================================================================
    // Step 5.1: UART configuration
    uartInit();
    
    // ========================================================================
    // Step 6: System timer configuration
    // we use Timer2 in 16 bit mode
    T2CONbits.TCKPS0 = 1; // prescaler dividing by 8 -> counting @ 8750kHz
    PR2 = 8749; // counting from 0 to 8749 -> interrupts @ 1kHz
    IPC1bits.T2IP = 4; // lower priority than the UART
    IEC0bits.T2IE = 1; // enable timer 2 interrupt
    T2CONbits.TON = 1; // start timer
    
    // ========================================================================
    // Step 7: Interrupt configuration
    INTCON2bits.GIE = 1; // enable interrupts
    
    // This loop is a central part of the windows operating system:
    // it does nothing, but takes up any time available
    while(1);
    
    // We can never get here, but the compiler complains like a teenager if
    // main is not of type int, so it has to be int, and have a return...
    return(0);
}

// Timer 2 interrupt service routine
// is called every 1ms
void __attribute__((__interrupt__, __auto_psv__)) _T2Interrupt(void)
{
    static unsigned int milliseconds = 0;
    //CAN_MESSAGE_T testMessage;
    
    calibrationTimerInt();
    
    milliseconds++;
    if(milliseconds == 500)
    {
        setLED(BLUE, ON);
    }
    else if(milliseconds == 1000)
    {
        setLED(BLUE, OFF);
        /*if(!PORTBbits.RB12) // stop 1 is pressed
        {
            testMessage.IDE = 0; // no extended addressing
            testMessage.SRR = 0; // no remote frame
            testMessage.SID = 0x0012; // axis control, axis 2
            testMessage.DLC = 4; // 4 Bytes of data
            testMessage.unused_null = 0; // there are two reserved 0-bits in there
            testMessage.RTR = 0; // not a remote frame
            testMessage.WORD0 = 0; // direction 0
            testMessage.WORD1 = 0x7FFF; // half speed
            canSendMessage(0, &testMessage);
            setLED(GREEN, ON);
        }
        else if(!PORTBbits.RB11) // stop 2 is pressed
        {
            testMessage.IDE = 0; // no extended addressing
            testMessage.SRR = 0; // no remote frame
            testMessage.SID = 0x0012; // axis control, axis 2
            testMessage.DLC = 4; // 4 Bytes of data
            testMessage.unused_null = 0; // there are two reserved 0-bits in there
            testMessage.RTR = 0; // not a remote frame
            testMessage.WORD0 = 1; // direction 1
            testMessage.WORD1 = 0x7FFF; // half speed
            canSendMessage(0, &testMessage);
            setLED(GREEN, ON);
        }
        else // no buttons pressed
        {
            testMessage.IDE = 0; // no extended addressing
            testMessage.SRR = 0; // no remote frame
            testMessage.SID = 0x0012; // axis control, axis 2
            testMessage.DLC = 4; // 4 Bytes of data
            testMessage.unused_null = 0; // there are two reserved 0-bits in there
            testMessage.RTR = 0; // not a remote frame
            testMessage.WORD0 = 0; // direction 0
            testMessage.WORD1 = 0; // stop
            canSendMessage(0, &testMessage);
            setLED(GREEN, OFF);
        }*/
        milliseconds = 0;
    }
    // end of timer 2 interrupt routine
    IFS0bits.T2IF = 0; // reset interrupt flag
}
