#include <xc.h>
#include "sw_config.h"
#include "led.h"
#include "pwm.h"
#include "qei.h"
#include "uart.h"

static struct {
    unsigned int bytesToSend;
    unsigned char* txBuffer;
} uartStatus;

void uartInit(void)
{
    U1BRG = 455; // 9,6k
    U1MODEbits.PDSEL = 0; // 8 data bits, no parity
    U1MODEbits.STSEL = 0; // 1 stop bit
    IPC3bits.U1TXIP = 5; // set priority level to 5 (arbitrary value)
    IPC2bits.U1RXIP = 5;
    U1MODEbits.UARTEN = 1; // enable the UART
    U1STAbits.UTXISEL1 = 1; // interrupt whenever the transmit buffer is empty
    U1STAbits.URXISEL = 0; // interrupt whenever a character is received
    U1STAbits.UTXEN = 1; // enable transmissions
    IEC0bits.U1RXIE = 1; // enable Rx interrupt
}

// Start sending a buffer
// buffer must be 0-terminated
int uartStartTx(unsigned char* buffer, unsigned int length)
{
    if(!uartStatus.bytesToSend) // uart is free
    {
        uartStatus.txBuffer = buffer;
        uartStatus.bytesToSend = length;
        U1STAbits.UTXEN = 1; // enable transmissions
        IEC0bits.U1TXIE = 1; // enable Tx interrupt
        return 0;
    }
    else return -1; // uart not free
}

// UART 1 Rx interrupt service routine
void __attribute__((__interrupt__, __auto_psv__)) _U1RXInterrupt(void)
{
    unsigned char receivedChar;
    
    receivedChar = U1RXREG; // read from UART
    switch(receivedChar)
    {
        case 0:
            pwmSetDutyCycle(0);
            setLED(GREEN, OFF);
            setLED(RED, ON);
            break;
        case 100:
            LATBbits.LATB14 = 1;
            pwmSetDutyCycle(5000);
            setLED(GREEN,ON);
            break;
        case 101:
            qeiSetSimpleHomingMode();
            LATBbits.LATB14 = 0;
            pwmSetDutyCycle(5000);
            setLED(YELLOW, ON);
            break;
        case 102:
            LATBbits.LATB14 = 0;
            pwmSetDutyCycle(5000);
            setLED(GREEN,ON);
            break;
    }
    
    IFS0bits.U1RXIF = 0; // reset interrupt flag
}

// UART 1 Tx interrupt service routine
void __attribute__((__interrupt__, __auto_psv__)) _U1TXInterrupt(void)
{
    while(!U1STAbits.UTXBF) // we fill the transmit buffer until it is full
    {
        // get a byte from the buffer
        if(uartStatus.bytesToSend) // end of buffer not reached
        {
            U1TXREG = *uartStatus.txBuffer;
            uartStatus.txBuffer++;
            uartStatus.bytesToSend--;
        }
        else // end of string, stop sending
        {
            IEC0bits.U1TXIE = 0; // disable Tx interrupt
            return;
        }
    }
    IFS0bits.U1TXIF = 0; // clear interrupt flag
}

void calibrationTimerInt(void)
{
    static unsigned int hundretMilliSeconds = 0; // for the 100ms updated - uart
    long int encoderValue;
    unsigned char messageBuffer[10];
    
    hundretMilliSeconds++;
    if(hundretMilliSeconds == 100)
    {
        encoderValue = qeiReadPosition();
        messageBuffer[0] = '*';
        messageBuffer[1] = '*';
        messageBuffer[2] = '*';
        messageBuffer[3] = '*';
        messageBuffer[4] = '*';
        messageBuffer[5] = '*';
        messageBuffer[6] = (unsigned char) (encoderValue >> 24);
        messageBuffer[7] = (unsigned char) (encoderValue >> 16);
        messageBuffer[8] = (unsigned char) (encoderValue >> 8);
        messageBuffer[9] = (unsigned char) (encoderValue);
        uartStartTx(messageBuffer, 10);
        
        hundretMilliSeconds = 0;
    }
}